use ark_ff::PrimeField;
use ark_poly::DenseUVPolynomial;
use ark_std::ops::Div;
use rs_merkle::algorithms::Sha256;
use rs_merkle::Hasher;
use tracing::{debug, info};

use crate::{algebra, error::KomodoError, fec};
use fri::{
    frida::{FridaBuilder, FridaCommitment},
    interpolation::interpolate_polynomials,
    rng::FriChallenger,
    utils::{to_evaluations, HasherExt, MerkleProof},
};

/// representation of a block of proven data.
///
/// this is a wrapper around a [`fec::Shard`] with some additional cryptographic
/// information that allows to prove the integrity of said shard.
#[derive(Clone, PartialEq)]
pub struct Block<F: PrimeField, H: Hasher> {
    pub shard: fec::Shard<F>,
    pub proof: MerkleProof<H>,
    pub commit: FridaCommitment<F, H>,
    position: usize,
}

/// # Panics
/// - see [`fri::utils::to_evaluations`] where each row in `bytes` is a `polynomial` and `n` is the
///   domain size
pub fn evaluate<F: PrimeField>(bytes: &[u8], k: usize, n: usize) -> Vec<Vec<F>> {
    debug!("splitting bytes into rows");
    let elements: Vec<F> = algebra::split_data_into_field_elements(bytes, k);
    let rows = elements.chunks(k).map(|c| c.to_vec()).collect::<Vec<_>>();
    info!(
        "data is composed of {} rows and {} elements",
        rows.len(),
        elements.len()
    );

    rows.into_iter()
        .map(|r| to_evaluations(r, n))
        .collect::<Vec<_>>()
}

#[inline]
pub fn transpose<F: Copy>(v: Vec<Vec<F>>) -> Vec<Vec<F>> {
    let mut cols: Vec<_> = Vec::<Vec<F>>::with_capacity(v[0].len());
    for i in 0..v[0].len() {
        cols.push((0..v.len()).map(|j| v[j][i]).collect());
    }
    cols
}

/// # Panics
/// - `n` must be a power of 2
pub fn encode<F: PrimeField>(
    bytes: &[u8],
    evaluations: Vec<Vec<F>>,
    k: usize,
) -> Vec<fec::Shard<F>> {
    let hash = Sha256::hash(bytes).to_vec();

    let n = evaluations[0].len();
    let w = F::get_root_of_unity(n as u64).unwrap();

    let t = transpose(evaluations);

    (0..n)
        .map(|i| fec::Shard {
            k: k as u32,
            linear_combination: (0..k).map(|j| w.pow([(i + j) as u64])).collect(),
            hash: hash.clone(),
            data: t[i].clone(),
            size: bytes.len(),
        })
        .collect::<Vec<_>>()
}

/// # Panics
/// - see [`FridaBuilder::new`] where the arguments are passed forward, as is
pub fn prove<const N: usize, F: PrimeField, H: Hasher, P>(
    evaluations: Vec<Vec<F>>,
    shards: Vec<fec::Shard<F>>,
    blowup_factor: usize,
    remainder_plus_one: usize,
    nb_queries: usize,
) -> Result<Vec<Block<F, H>>, KomodoError>
where
    P: DenseUVPolynomial<F>,
    for<'a, 'b> &'a P: Div<&'b P, Output = P>,
    <H as rs_merkle::Hasher>::Hash: AsRef<[u8]>,
{
    assert!(blowup_factor > 1, "BF should be strictly greater than 1");

    assert!(
        shards[0].k > (remainder_plus_one as u32),
        "degree bound should be greater than remainder, found {} and {}",
        shards[0].k,
        remainder_plus_one,
    );

    assert!(
        (evaluations[0].len() as f64).log(N as f64) % 1.0 == 0.0
            && (shards[0].k as f64).log(N as f64) % 1.0 == 0.0,
        "bad parameters"
    );

    let builder = FridaBuilder::<F, H>::new::<N, _>(
        &evaluations,
        FriChallenger::<H>::default(),
        blowup_factor,
        remainder_plus_one,
        nb_queries,
    );

    let commit = FridaCommitment::from(builder.clone());

    Ok(shards
        .iter()
        .enumerate()
        .map(|(i, s)| Block {
            shard: s.clone(),
            proof: builder.prove_shards(&[i]),
            commit: commit.clone(),
            position: i,
        })
        .collect())
}

/// # Panics
/// - see [`FridaCommitment::verify`] where `degree_bound` is the value of `k` stored in each
///   [`Block`]
pub fn verify<const N: usize, F: PrimeField, H: Hasher, P>(
    block: Block<F, H>,
    domain_size: usize,
    nb_queries: usize,
) -> Result<(), KomodoError>
where
    P: DenseUVPolynomial<F>,
    for<'a, 'b> &'a P: Div<&'b P, Output = P>,
    <H as rs_merkle::Hasher>::Hash: AsRef<[u8]>,
{
    block
        .commit
        .verify::<N, _>(
            FriChallenger::<H>::default(),
            nb_queries,
            block.shard.k as usize,
            domain_size,
        )
        .unwrap();

    assert!(block.proof.verify(
        block.commit.tree_root(),
        &[block.position],
        &[H::hash_item(&block.shard.data)],
        domain_size,
    ));

    Ok(())
}

/// # Panics
/// - see [`interpolate_polynomials`] where `evaluations` is each one of the transposed shards and
///   the positions are the ones stored in each [`Block`]
pub fn decode<F: PrimeField, H: Hasher>(blocks: Vec<Block<F, H>>, n: usize) -> Vec<u8> {
    let w = F::get_root_of_unity(n as u64).unwrap();

    let t_shards = transpose(blocks.iter().map(|b| b.shard.data.clone()).collect());
    let positions = blocks
        .iter()
        .map(|b| w.pow([b.position as u64]))
        .collect::<Vec<_>>();
    let source_shards = interpolate_polynomials(&t_shards, &positions)
        .into_iter()
        .flatten()
        .collect::<Vec<_>>();

    let mut bytes = algebra::merge_elements_into_bytes(&source_shards);
    bytes.resize(blocks[0].shard.size, 0);
    bytes
}
